//
//  HDKeychain.swift
//
//  Copyright © 2018 Kishikawa Katsumi
//  Copyright © 2018 BitcoinKit developers
//
//  Permission is hereby granted, free of charge, to any person obtaining a copy
//  of this software and associated documentation files (the "Software"), to deal
//  in the Software without restriction, including without limitation the rights
//  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  copies of the Software, and to permit persons to whom the Software is
//  furnished to do so, subject to the following conditions:
//
//  The above copyright notice and this permission notice shall be included in
//  all copies or substantial portions of the Software.
//
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  THE SOFTWARE.
//

import Foundation

public final class HDKeychain {
  private let rootKey: HDPrivateKey

  public init(rootKey: HDPrivateKey) {
    self.rootKey = rootKey
  }

  public convenience init(seed: Data) {
    self.init(rootKey: HDPrivateKey(seed: seed))
  }

  /// Parses the BIP32 path and derives the chain of keychains accordingly.
  /// Path syntax: (m?/)?([0-9]+'?(/[0-9]+'?)*)?
  /// The following paths are valid:
  ///
  /// "" (root key)
  /// "m" (root key)
  /// "/" (root key)
  /// "m/0'" (hardened child #0 of the root key)
  /// "/0'" (hardened child #0 of the root key)
  /// "0'" (hardened child #0 of the root key)
  /// "m/44'/1'/2'" (BIP44 testnet account #2)
  /// "/44'/1'/2'" (BIP44 testnet account #2)
  /// "44'/1'/2'" (BIP44 testnet account #2)
  ///
  /// The following paths are invalid:
  ///
  /// "m / 0 / 1" (contains spaces)
  /// "m/b/c" (alphabetical characters instead of numerical indexes)
  /// "m/1.2^3" (contains illegal characters)
  public func derivedKey(path: String) throws -> HDPrivateKey {
    var key = rootKey

    var path = path
    if path == "m" || path == "/" || path == "" {
      return key
    }
    if path.contains("m/") {
      path = String(path.dropFirst(2))
    }
    for chunk in path.split(separator: "/") {
      var hardened = false
      var indexText = chunk
      if chunk.contains("'") {
        hardened = true
        indexText = indexText.dropLast()
      }
      guard let index = UInt32(indexText) else {
        fatalError("invalid path")
      }
      key = try key.derived(at: index, hardened: hardened)
    }
    return key
  }
}
