//
//  Mnemonic.swift
//
//  Copyright © 2018 Kishikawa Katsumi
//  Copyright © 2018 BitcoinKit developers
//
//  Permission is hereby granted, free of charge, to any person obtaining a copy
//  of this software and associated documentation files (the "Software"), to deal
//  in the Software without restriction, including without limitation the rights
//  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  copies of the Software, and to permit persons to whom the Software is
//  furnished to do so, subject to the following conditions:
//
//  The above copyright notice and this permission notice shall be included in
//  all copies or substantial portions of the Software.
//
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  THE SOFTWARE.
//

import Foundation

public struct Mnemonic {
  public enum Strength: Int {
    case `default` = 128
    case low = 160
    case medium = 192
    case high = 224
    case veryHigh = 256
  }

  public enum Language {
    case english
    case japanese
    case korean
    case spanish
    case simplifiedChinese
    case traditionalChinese
    case french
    case italian
  }

  public static func generate(strength: Strength = .default, language: Language = .english) throws -> [String] {
    let byteCount = strength.rawValue / 8
    var bytes = Data(count: byteCount)
    let status = bytes.withUnsafeMutableBytes { SecRandomCopyBytes(kSecRandomDefault, byteCount, $0) }
    guard status == errSecSuccess else { throw MnemonicError.randomBytesError }
    return generate(entropy: bytes, language: language)
  }

  internal static func generate(entropy: Data, language: Language = .english) -> [String] {
    let list = wordList(for: language)
    var bin = String(entropy.flatMap { ("00000000" + String($0, radix: 2)).suffix(8) })

    let hash = _Hash.sha256(entropy)
    let bits = entropy.count * 8
    let cs = bits / 32

    let hashbits = String(hash.flatMap { ("00000000" + String($0, radix: 2)).suffix(8) })
    let checksum = String(hashbits.prefix(cs))
    bin += checksum

    var mnemonic = [String]()
    for i in 0 ..< (bin.count / 11) {
      let wi = Int(bin[bin.index(bin.startIndex, offsetBy: i * 11) ..< bin.index(bin.startIndex, offsetBy: (i + 1) * 11)], radix: 2)!
      mnemonic.append(String(list[wi]))
    }
    return mnemonic
  }

  public static func seed(mnemonic m: [String], passphrase: String = "") -> Data {
    let mnemonic = m.joined(separator: " ").decomposedStringWithCompatibilityMapping.data(using: .utf8)!
    let salt = ("mnemonic" + passphrase).decomposedStringWithCompatibilityMapping.data(using: .utf8)!
    let seed = _Key.deriveKey(mnemonic, salt: salt, iterations: 2048, keyLength: 64)
    return seed
  }

  private static func wordList(for language: Language) -> [String.SubSequence] {
    switch language {
    case .english:
      return WordList.english
    case .japanese:
      return WordList.japanese
    case .korean:
      return WordList.korean
    case .spanish:
      return WordList.spanish
    case .simplifiedChinese:
      return WordList.simplifiedChinese
    case .traditionalChinese:
      return WordList.traditionalChinese
    case .french:
      return WordList.french
    case .italian:
      return WordList.italian
    }
  }
}

public enum MnemonicError: Error {
  case randomBytesError
}
